@@ -1,6 +1,6 @@
 /* -*- Mode: C; tab-width: 8; indent-tabs-mode: t; c-basic-offset: 8 -*- */
 /*  GMime
- *  Copyright (C) 2000-2010 Jeffrey Stedfast
+ *  Copyright (C) 2000-2009 Jeffrey Stedfast
  *
  *  This library is free software; you can redistribute it and/or
  *  modify it under the terms of the GNU Lesser General Public License
@@ -103,15 +103,6 @@ static int gpg_export_keys (GMimeCipherContext *ctx, GPtrArray *keys,
 static GMimeCipherContextClass *parent_class = NULL;
 
 
-#if GLIB_MAJOR_VERSION == 2 && GLIB_MINOR_VERSION < 18
-static void
-g_set_error_literal (GError **err, GQuark domain, gint code, const gchar *message)
-{
-	g_set_error (err, domain, code, "%s", message);
-}
-#endif
-
-
 GType
 g_mime_gpg_context_get_type (void)
 {
@@ -261,6 +252,7 @@ struct _GpgCtx {
 	
 	char *path;
 	char *userid;
+	char *sigfile;
 	GPtrArray *recipients;
 	GMimeCipherHash hash;
 	
@@ -268,7 +260,7 @@ struct _GpgCtx {
 	int stdout_fd;
 	int stderr_fd;
 	int status_fd;
-	int secret_fd;  /* used for sign/decrypt/verify */
+	int passwd_fd;  /* only needed for sign/decrypt */
 	
 	/* status-fd buffer */
 	char *statusbuf;
@@ -278,7 +270,6 @@ struct _GpgCtx {
 	char *need_id;
 	char *passwd;
 	
-	GMimeStream *sigstream;
 	GMimeStream *istream;
 	GMimeStream *ostream;
 	
@@ -335,6 +326,7 @@ gpg_ctx_new (GMimeSession *session, const char *path)
 	
 	gpg->path = g_strdup (path);
 	gpg->userid = NULL;
+	gpg->sigfile = NULL;
 	gpg->recipients = NULL;
 	gpg->hash = GMIME_CIPHER_HASH_DEFAULT;
 	gpg->always_trust = FALSE;
@@ -344,7 +336,7 @@ gpg_ctx_new (GMimeSession *session, const char *path)
 	gpg->stdout_fd = -1;
 	gpg->stderr_fd = -1;
 	gpg->status_fd = -1;
-	gpg->secret_fd = -1;
+	gpg->passwd_fd = -1;
 	
 	gpg->statusbuf = g_malloc (128);
 	gpg->statusptr = gpg->statusbuf;
@@ -366,7 +358,6 @@ gpg_ctx_new (GMimeSession *session, const char *path)
 	gpg->signers = NULL;
 	gpg->signer = (GMimeSigner *) &gpg->signers;
 	
-	gpg->sigstream = NULL;
 	gpg->istream = NULL;
 	gpg->ostream = NULL;
 	
@@ -436,6 +427,13 @@ gpg_ctx_add_recipient (struct _GpgCtx *gpg, const char *keyid)
 	g_ptr_array_add (gpg->recipients, g_strdup (keyid));
 }
 
+static void
+gpg_ctx_set_sigfile (struct _GpgCtx *gpg, const char *sigfile)
+{
+	g_free (gpg->sigfile);
+	gpg->sigfile = g_strdup (sigfile);
+}
+
 static void
 gpg_ctx_set_armor (struct _GpgCtx *gpg, gboolean armor)
 {
@@ -461,15 +459,6 @@ gpg_ctx_set_ostream (struct _GpgCtx *gpg, GMimeStream *ostream)
 	gpg->seen_eof1 = FALSE;
 }
 
-static void
-gpg_ctx_set_sigstream (struct _GpgCtx *gpg, GMimeStream *sigstream)
-{
-	g_object_ref (sigstream);
-	if (gpg->sigstream)
-		g_object_unref (gpg->sigstream);
-	gpg->sigstream = sigstream;
-}
-
 static const char *
 gpg_ctx_get_diagnostics (struct _GpgCtx *gpg)
 {
@@ -486,7 +475,7 @@ static void
 gpg_ctx_free (struct _GpgCtx *gpg)
 {
 	GMimeSigner *signer, *next;
-	guint i;
+	int i;
 	
 	if (gpg->session)
 		g_object_unref (gpg->session);
@@ -497,6 +486,8 @@ gpg_ctx_free (struct _GpgCtx *gpg)
 	
 	g_free (gpg->userid);
 	
+	g_free (gpg->sigfile);
+	
 	if (gpg->recipients) {
 		for (i = 0; i < gpg->recipients->len; i++)
 			g_free (gpg->recipients->pdata[i]);
@@ -512,8 +503,8 @@ gpg_ctx_free (struct _GpgCtx *gpg)
 		close (gpg->stderr_fd);
 	if (gpg->status_fd != -1)
 		close (gpg->status_fd);
-	if (gpg->secret_fd != -1)
-		close (gpg->secret_fd);
+	if (gpg->passwd_fd != -1)
+		close (gpg->passwd_fd);
 	
 	g_free (gpg->statusbuf);
 	
@@ -524,9 +515,6 @@ gpg_ctx_free (struct _GpgCtx *gpg)
 		g_free (gpg->passwd);
 	}
 	
-	if (gpg->sigstream)
-		g_object_unref (gpg->sigstream);
-	
 	if (gpg->istream)
 		g_object_unref (gpg->istream);
 	
@@ -572,134 +560,118 @@ gpg_hash_str (GMimeCipherHash hash)
 	}
 }
 
-static char **
-gpg_ctx_get_argv (struct _GpgCtx *gpg, int status_fd, int secret_fd, char ***strv)
+static GPtrArray *
+gpg_ctx_get_argv (struct _GpgCtx *gpg, int status_fd, char **sfd, int passwd_fd, char **pfd)
 {
 	const char *hash_str;
-	char **argv, *buf;
-	GPtrArray *args;
-	int v = 0;
-	guint i;
+	GPtrArray *argv;
+	char *buf;
+	int i;
 	
-	*strv = g_new (char *, 3);
+	argv = g_ptr_array_new ();
+	g_ptr_array_add (argv, "gpg");
 	
-	args = g_ptr_array_new ();
-	g_ptr_array_add (args, "gpg");
-	
-	g_ptr_array_add (args, "--verbose");
-	g_ptr_array_add (args, "--no-secmem-warning");
-	g_ptr_array_add (args, "--no-greeting");
-	g_ptr_array_add (args, "--no-tty");
-	
-	if (!gpg->need_passwd) {
+	g_ptr_array_add (argv, "--verbose");
+	g_ptr_array_add (argv, "--no-secmem-warning");
+	g_ptr_array_add (argv, "--no-greeting");
+	g_ptr_array_add (argv, "--no-tty");
+	if (passwd_fd == -1) {
 		/* only use batch mode if we don't intend on using the
-                   interactive --command-fd option to send it the
-                   user's password */
-		g_ptr_array_add (args, "--batch");
-		g_ptr_array_add (args, "--yes");
+                   interactive --command-fd option */
+		g_ptr_array_add (argv, "--batch");
+		g_ptr_array_add (argv, "--yes");
 	}
 	
-	g_ptr_array_add (args, "--charset=UTF-8");
+	g_ptr_array_add (argv, "--charset=UTF-8");
 	
-	(*strv)[v++] = buf = g_strdup_printf ("--status-fd=%d", status_fd);
-	g_ptr_array_add (args, buf);
+	*sfd = buf = g_strdup_printf ("--status-fd=%d", status_fd);
+	g_ptr_array_add (argv, buf);
 	
-	if (gpg->need_passwd) {
-		(*strv)[v++] = buf = g_strdup_printf ("--command-fd=%d", secret_fd);
-		g_ptr_array_add (args, buf);
+	if (passwd_fd != -1) {
+		*pfd = buf = g_strdup_printf ("--command-fd=%d", passwd_fd);
+		g_ptr_array_add (argv, buf);
 	}
 	
 	switch (gpg->mode) {
 	case GPG_CTX_MODE_SIGN:
-		g_ptr_array_add (args, "--sign");
-		g_ptr_array_add (args, "--detach");
+		g_ptr_array_add (argv, "--sign");
+		g_ptr_array_add (argv, "--detach");
 		if (gpg->armor)
-			g_ptr_array_add (args, "--armor");
+			g_ptr_array_add (argv, "--armor");
 		hash_str = gpg_hash_str (gpg->hash);
 		if (hash_str)
-			g_ptr_array_add (args, (char *) hash_str);
+			g_ptr_array_add (argv, (char *) hash_str);
 		if (gpg->userid) {
-			g_ptr_array_add (args, "-u");
-			g_ptr_array_add (args, (char *) gpg->userid);
+			g_ptr_array_add (argv, "-u");
+			g_ptr_array_add (argv, (char *) gpg->userid);
 		}
-		g_ptr_array_add (args, "--output");
-		g_ptr_array_add (args, "-");
+		g_ptr_array_add (argv, "--output");
+		g_ptr_array_add (argv, "-");
 		break;
 	case GPG_CTX_MODE_VERIFY:
 		if (!g_mime_session_is_online (gpg->session)) {
 			/* this is a deprecated flag to gpg since 1.0.7 */
-			/*g_ptr_array_add (args, "--no-auto-key-retrieve");*/
-			g_ptr_array_add (args, "--keyserver-options");
-			g_ptr_array_add (args, "no-auto-key-retrieve");
+			/*g_ptr_array_add (argv, "--no-auto-key-retrieve");*/
+			g_ptr_array_add (argv, "--keyserver-options");
+			g_ptr_array_add (argv, "no-auto-key-retrieve");
 		}
-		
-		g_ptr_array_add (args, "--enable-special-filenames");
-		g_ptr_array_add (args, "--verify");
-		g_ptr_array_add (args, "--");
-		
-		/* signature stream must come first */
-		(*strv)[v++] = buf = g_strdup_printf ("-&%d", secret_fd);
-		g_ptr_array_add (args, buf);
-		
-		/* followed by the content stream (in this case, stdin) */
-		g_ptr_array_add (args, "-");
+		g_ptr_array_add (argv, "--verify");
+		if (gpg->sigfile)
+			g_ptr_array_add (argv, gpg->sigfile);
+		g_ptr_array_add (argv, "-");
 		break;
 	case GPG_CTX_MODE_SIGN_ENCRYPT:
-		g_ptr_array_add (args, "--sign");
+		g_ptr_array_add (argv,  "--sign");
 		
 		/* fall thru... */
 	case GPG_CTX_MODE_ENCRYPT:
-		g_ptr_array_add (args, "--encrypt");
+		g_ptr_array_add (argv,  "--encrypt");
 		
 		if (gpg->armor)
-			g_ptr_array_add (args, "--armor");
+			g_ptr_array_add (argv, "--armor");
 		
 		if (gpg->always_trust)
-			g_ptr_array_add (args, "--always-trust");
+			g_ptr_array_add (argv, "--always-trust");
 		
 		if (gpg->userid) {
-			g_ptr_array_add (args, "-u");
-			g_ptr_array_add (args, (char *) gpg->userid);
+			g_ptr_array_add (argv, "-u");
+			g_ptr_array_add (argv, (char *) gpg->userid);
 		}
 		
 		if (gpg->recipients) {
 			for (i = 0; i < gpg->recipients->len; i++) {
-				g_ptr_array_add (args, "-r");
-				g_ptr_array_add (args, gpg->recipients->pdata[i]);
+				g_ptr_array_add (argv, "-r");
+				g_ptr_array_add (argv, gpg->recipients->pdata[i]);
 			}
 		}
-		g_ptr_array_add (args, "--output");
-		g_ptr_array_add (args, "-");
+		g_ptr_array_add (argv, "--output");
+		g_ptr_array_add (argv, "-");
 		break;
 	case GPG_CTX_MODE_DECRYPT:
-		g_ptr_array_add (args, "--decrypt");
-		g_ptr_array_add (args, "--output");
-		g_ptr_array_add (args, "-");
+		g_ptr_array_add (argv, "--decrypt");
+		g_ptr_array_add (argv, "--output");
+		g_ptr_array_add (argv, "-");
 		break;
 	case GPG_CTX_MODE_IMPORT:
-		g_ptr_array_add (args, "--import");
-		g_ptr_array_add (args, "-");
+		g_ptr_array_add (argv, "--import");
+		g_ptr_array_add (argv, "-");
 		break;
 	case GPG_CTX_MODE_EXPORT:
 		if (gpg->armor)
-			g_ptr_array_add (args, "--armor");
-		g_ptr_array_add (args, "--export");
+			g_ptr_array_add (argv, "--armor");
+		g_ptr_array_add (argv, "--export");
 		for (i = 0; i < gpg->recipients->len; i++)
-			g_ptr_array_add (args, gpg->recipients->pdata[i]);
+			g_ptr_array_add (argv, gpg->recipients->pdata[i]);
 		break;
 	}
 	
 #if d(!)0
-	for (i = 0; i < args->len; i++)
-		printf ("%s ", (char *) args->pdata[i]);
+	for (i = 0; i < argv->len; i++)
+		printf ("%s ", (char *) argv->pdata[i]);
 	printf ("\n");
 #endif
 	
-	g_ptr_array_add (args, NULL);
-	(*strv)[v] = NULL;
-	
-	argv = (char **) args->pdata;
-	g_ptr_array_free (args, FALSE);
+	g_ptr_array_add (argv, NULL);
 	
 	return argv;
 }
@@ -707,27 +679,28 @@ gpg_ctx_get_argv (struct _GpgCtx *gpg, int status_fd, int secret_fd, char ***str
 static int
 gpg_ctx_op_start (struct _GpgCtx *gpg)
 {
+	char *status_fd = NULL, *passwd_fd = NULL;
 	int i, maxfd, errnosave, fds[10];
-	char **argv, **strv = NULL;
+	GPtrArray *argv;
 	int flags;
 	
 	for (i = 0; i < 10; i++)
 		fds[i] = -1;
 	
-	maxfd = (gpg->need_passwd || gpg->sigstream) ? 10 : 8;
+	maxfd = gpg->need_passwd ? 10 : 8;
 	for (i = 0; i < maxfd; i += 2) {
 		if (pipe (fds + i) == -1)
 			goto exception;
 	}
 	
-	argv = gpg_ctx_get_argv (gpg, fds[7], fds[8], &strv);
+	argv = gpg_ctx_get_argv (gpg, fds[7], &status_fd, fds[8], &passwd_fd);
 	
 	if (!(gpg->pid = fork ())) {
 		/* child process */
 		
-		if ((dup2 (fds[0], STDIN_FILENO) < 0) ||
-		    (dup2 (fds[3], STDOUT_FILENO) < 0) ||
-		    (dup2 (fds[5], STDERR_FILENO) < 0)) {
+		if ((dup2 (fds[0], STDIN_FILENO) < 0 ) ||
+		    (dup2 (fds[3], STDOUT_FILENO) < 0 ) ||
+		    (dup2 (fds[5], STDERR_FILENO) < 0 )) {
 			_exit (255);
 		}
 		
@@ -744,19 +717,20 @@ gpg_ctx_op_start (struct _GpgCtx *gpg)
 		}
 		
 		/* run gpg */
-		execvp (gpg->path, argv);
+		execvp (gpg->path, (char **) argv->pdata);
 		_exit (255);
 	} else if (gpg->pid < 0) {
-		g_strfreev (strv);
-		g_free (argv);
+		g_ptr_array_free (argv, TRUE);
+		g_free (status_fd);
+		g_free (passwd_fd);
 		goto exception;
 	}
 	
-	/* parent process */
-	
-	g_strfreev (strv);
-	g_free (argv);
+	g_ptr_array_free (argv, TRUE);
+	g_free (status_fd);
+	g_free (passwd_fd);
 	
+	/* Parent */
 	close (fds[0]);
 	gpg->stdin_fd = fds[1];
 	gpg->stdout_fd = fds[2];
@@ -765,12 +739,11 @@ gpg_ctx_op_start (struct _GpgCtx *gpg)
 	close (fds[5]);
 	gpg->status_fd = fds[6];
 	close (fds[7]);
-	
-	if (fds[8] != -1) {
-		flags = (flags = fcntl (fds[9], F_GETFL)) == -1 ? O_WRONLY : flags;
-		fcntl (fds[9], F_SETFL, flags | O_NONBLOCK);
-		gpg->secret_fd = fds[9];
+	if (gpg->need_passwd) {
 		close (fds[8]);
+		gpg->passwd_fd = fds[9];
+		flags = (flags = fcntl (gpg->passwd_fd, F_GETFL)) == -1 ? O_WRONLY : flags;
+		fcntl (gpg->passwd_fd, F_SETFL, flags | O_NONBLOCK);
 	}
 	
 	flags = (flags = fcntl (gpg->stdin_fd, F_GETFL)) == -1 ? O_WRONLY : flags;
@@ -791,7 +764,7 @@ gpg_ctx_op_start (struct _GpgCtx *gpg)
 	
 	errnosave = errno;
 	
-	for (i = 0; i < maxfd; i++) {
+	for (i = 0; i < 10; i++) {
 		if (fds[i] != -1)
 			close (fds[i]);
 	}
@@ -820,39 +793,11 @@ next_token (char *in, char **token)
 		inptr++;
 	
 	if (token)
-		*token = g_strndup (start, (size_t) (inptr - start));
+		*token = g_strndup (start, inptr - start);
 	
 	return inptr;
 }
 
-/**
- * gpg_ctx_add_signer:
- * @gpg: GnuPG context
- * @info: a string with the signer's info
- *
- * Parses GOODSIG, BADSIG, EXPSIG, EXPKEYSIG, and REVKEYSIG status messages
- * into a newly allocated #GMimeSigner, adding it to @gpg's signer list.
- *
- * Returns: the newly added signer for the caller to add more info to.
- **/
-static GMimeSigner *
-gpg_ctx_add_signer (struct _GpgCtx *gpg, char *info)
-{
-	GMimeSigner *signer;
-	
-	signer = g_mime_signer_new ();
-	gpg->signer->next = signer;
-	gpg->signer = signer;
-	
-	/* get the key id of the signer */
-	info = next_token (info, &signer->keyid);
-	
-	/* the rest of the string is the signer's name */
-	signer->name = g_strdup (info);
-	
-	return signer;
-}
-
 static void
 gpg_ctx_parse_signer_info (struct _GpgCtx *gpg, char *status)
 {
@@ -861,25 +806,33 @@ gpg_ctx_parse_signer_info (struct _GpgCtx *gpg, char *status)
 	if (!strncmp (status, "SIG_ID ", 7)) {
 		/* not sure if this contains anything we care about... */
 	} else if (!strncmp (status, "GOODSIG ", 8)) {
-		signer = gpg_ctx_add_signer (gpg, status + 8);
-		signer->status = GMIME_SIGNER_STATUS_GOOD;
 		gpg->goodsig = TRUE;
+		status += 8;
+		
+		signer = g_mime_signer_new ();
+		signer->status = GMIME_SIGNER_STATUS_GOOD;
+		gpg->signer->next = signer;
+		gpg->signer = signer;
+		
+		/* get the key id of the signer */
+		status = next_token (status, &signer->keyid);
+		
+		/* the rest of the string is the signer's name */
+		signer->name = g_strdup (status);
 	} else if (!strncmp (status, "BADSIG ", 7)) {
-		signer = gpg_ctx_add_signer (gpg, status + 7);
-		signer->status = GMIME_SIGNER_STATUS_BAD;
 		gpg->badsig = TRUE;
-	} else if (!strncmp (status, "EXPSIG ", 7)) {
-		signer = gpg_ctx_add_signer (gpg, status + 7);
-		signer->status = GMIME_SIGNER_STATUS_ERROR;
-		signer->errors = GMIME_SIGNER_ERROR_EXPSIG;
-	} else if (!strncmp (status, "EXPKEYSIG ", 10)) {
-		signer = gpg_ctx_add_signer (gpg, status + 10);
-		signer->status = GMIME_SIGNER_STATUS_ERROR;
-		signer->errors = GMIME_SIGNER_ERROR_EXPKEYSIG;
-	} else if (!strncmp (status, "REVKEYSIG ", 10)) {
-		signer = gpg_ctx_add_signer (gpg, status + 10);
-		signer->status = GMIME_SIGNER_STATUS_ERROR;
-		signer->errors = GMIME_SIGNER_ERROR_REVKEYSIG;
+		status += 7;
+		
+		signer = g_mime_signer_new ();
+		signer->status = GMIME_SIGNER_STATUS_BAD;
+		gpg->signer->next = signer;
+		gpg->signer = signer;
+		
+		/* get the key id of the signer */
+		status = next_token (status, &signer->keyid);
+		
+		/* the rest of the string is the signer's name */
+		signer->name = g_strdup (status);
 	} else if (!strncmp (status, "ERRSIG ", 7)) {
 		/* Note: NO_PUBKEY often comes after an ERRSIG */
 		gpg->errsig = TRUE;
@@ -911,6 +864,13 @@ gpg_ctx_parse_signer_info (struct _GpgCtx *gpg, char *status)
 		/* the only token is the keyid, but we've already got it */
 		gpg->signer->errors |= GMIME_SIGNER_ERROR_NO_PUBKEY;
 		gpg->nopubkey = TRUE;
+	} else if (!strncmp (status, "EXPSIG", 6)) {
+		/* FIXME: see what else we can glean from this... */
+		gpg->signer->errors |= GMIME_SIGNER_ERROR_EXPSIG;
+	} else if (!strncmp (status, "EXPKEYSIG", 9)) {
+		gpg->signer->errors |= GMIME_SIGNER_ERROR_EXPKEYSIG;
+	} else if (!strncmp (status, "REVKEYSIG", 9)) {
+		gpg->signer->errors |= GMIME_SIGNER_ERROR_REVKEYSIG;
 	} else if (!strncmp (status, "VALIDSIG ", 9)) {
 		char *inend;
 		
@@ -1006,8 +966,8 @@ gpg_ctx_parse_status (struct _GpgCtx *gpg, GError **err)
 		
 		status = next_token (status, &hint);
 		if (!hint) {
-			g_set_error_literal (err, GMIME_ERROR, GMIME_ERROR_PARSE_ERROR,
-					     _("Failed to parse gpg userid hint."));
+			g_set_error (err, GMIME_ERROR, GMIME_ERROR_PARSE_ERROR,
+				     _("Failed to parse gpg userid hint."));
 			return -1;
 		}
 		
@@ -1030,8 +990,8 @@ gpg_ctx_parse_status (struct _GpgCtx *gpg, GError **err)
 		
 		status = next_token (status, &userid);
 		if (!userid) {
-			g_set_error_literal (err, GMIME_ERROR, GMIME_ERROR_PARSE_ERROR,
-					     _("Failed to parse gpg passphrase request."));
+			g_set_error (err, GMIME_ERROR, GMIME_ERROR_PARSE_ERROR,
+				     _("Failed to parse gpg passphrase request."));
 			return -1;
 		}
 		
@@ -1044,8 +1004,8 @@ gpg_ctx_parse_status (struct _GpgCtx *gpg, GError **err)
 		
 		status = next_token (status, &userid);
 		if (!userid) {
-			g_set_error_literal (err, GMIME_ERROR, GMIME_ERROR_PARSE_ERROR,
-					     _("Failed to parse gpg passphrase request."));
+			g_set_error (err, GMIME_ERROR, GMIME_ERROR_PARSE_ERROR,
+				     _("Failed to parse gpg passphrase request."));
 			return -1;
 		}
 		
@@ -1095,7 +1055,7 @@ gpg_ctx_parse_status (struct _GpgCtx *gpg, GError **err)
 			gpg->send_passwd = TRUE;
 		} else {
 			if (err && *err == NULL)
-				g_set_error_literal (err, GMIME_ERROR, ECANCELED, _("Canceled."));
+				g_set_error (err, GMIME_ERROR, ECANCELED, _("Canceled."));
 			
 			g_free (prompt);
 			
@@ -1109,8 +1069,8 @@ gpg_ctx_parse_status (struct _GpgCtx *gpg, GError **err)
 		g_mime_session_forget_passwd (gpg->session, gpg->userid, NULL);
 		
 		if (gpg->bad_passwds == 3) {
-			g_set_error_literal (err, GMIME_ERROR, GMIME_ERROR_BAD_PASSWORD,
-					     _("Failed to unlock secret key: 3 bad passphrases given."));
+			g_set_error (err, GMIME_ERROR, GMIME_ERROR_BAD_PASSWORD,
+				     _("Failed to unlock secret key: 3 bad passphrases given."));
 			return -1;
 		}
 	} else if (!strncmp (status, "UNEXPECTED ", 11)) {
@@ -1134,9 +1094,11 @@ gpg_ctx_parse_status (struct _GpgCtx *gpg, GError **err)
 		
 		diagnostics = gpg_ctx_get_diagnostics (gpg);
 		if (diagnostics && *diagnostics)
-			g_set_error_literal (err, GMIME_ERROR, GMIME_ERROR_GENERAL, diagnostics);
+			g_set_error (err, GMIME_ERROR, GMIME_ERROR_GENERAL,
+				     "%s", diagnostics);
 		else
-			g_set_error_literal (err, GMIME_ERROR, GMIME_ERROR_GENERAL, _("No data provided"));
+			g_set_error (err, GMIME_ERROR, GMIME_ERROR_GENERAL,
+				     _("No data provided"));
 		
 		gpg->nodata = TRUE;
 		
@@ -1180,8 +1142,8 @@ gpg_ctx_parse_status (struct _GpgCtx *gpg, GError **err)
 			} else if (!strncmp (status, "END_ENCRYPTION", 14)) {
 				/* nothing to do, but we know the end is near? */
 			} else if (!strncmp (status, "NO_RECP", 7)) {
-				g_set_error_literal (err, GMIME_ERROR, GMIME_ERROR_NO_VALID_RECIPIENTS,
-						     _("Failed to encrypt: No valid recipients specified."));
+				g_set_error (err, GMIME_ERROR, GMIME_ERROR_NO_VALID_RECIPIENTS,
+					     _("Failed to encrypt: No valid recipients specified."));
 				return -1;
 			}
 			break;
@@ -1279,7 +1241,7 @@ enum {
 	GPG_STDOUT_FD,
 	GPG_STDERR_FD,
 	GPG_STATUS_FD,
-	GPG_SECRET_FD,
+	GPG_PASSWD_FD,
 	GPG_N_FDS
 };
 
@@ -1391,8 +1353,8 @@ gpg_ctx_op_step (struct _GpgCtx *gpg, GError **err)
 	pfds[GPG_STDIN_FD].fd = gpg->stdin_fd;
 	pfds[GPG_STDIN_FD].events = POLLOUT;
 	
-	pfds[GPG_SECRET_FD].fd = gpg->secret_fd;
-	pfds[GPG_SECRET_FD].events = POLLOUT;
+	pfds[GPG_PASSWD_FD].fd = gpg->passwd_fd;
+	pfds[GPG_PASSWD_FD].events = POLLOUT;
 	
 	do {
 		for (n = 0; n < GPG_N_FDS; n++)
@@ -1429,7 +1391,7 @@ gpg_ctx_op_step (struct _GpgCtx *gpg, GError **err)
 			goto exception;
 		
 		if (nread > 0) {
-			status_backup (gpg, buffer, (size_t) nread);
+			status_backup (gpg, buffer, nread);
 			if (gpg_ctx_parse_status (gpg, err) == -1)
 				return -1;
 		} else {
@@ -1446,7 +1408,6 @@ gpg_ctx_op_step (struct _GpgCtx *gpg, GError **err)
 		do {
 			nread = read (gpg->stdout_fd, buffer, sizeof (buffer));
 		} while (nread == -1 && (errno == EINTR || errno == EAGAIN));
-		
 		if (nread == -1)
 			goto exception;
 		
@@ -1467,7 +1428,6 @@ gpg_ctx_op_step (struct _GpgCtx *gpg, GError **err)
 		do {
 			nread = read (gpg->stderr_fd, buffer, sizeof (buffer));
 		} while (nread == -1 && (errno == EINTR || errno == EAGAIN));
-		
 		if (nread == -1)
 			goto exception;
 		
@@ -1478,60 +1438,32 @@ gpg_ctx_op_step (struct _GpgCtx *gpg, GError **err)
 		}
 	}
 	
-	if ((pfds[GPG_SECRET_FD].revents & (POLLOUT | POLLHUP))) {
-		if (gpg->need_passwd && gpg->send_passwd) {
-			size_t n, nwritten = 0;
-			ssize_t w;
-			
-			d(printf ("sending gpg our passphrase...\n"));
-			
-			/* send the passphrase to gpg */
-			n = strlen (gpg->passwd);
+	if ((pfds[GPG_PASSWD_FD].revents & (POLLOUT | POLLHUP)) && gpg->need_passwd && gpg->send_passwd) {
+		ssize_t w, nwritten = 0;
+		size_t n;
+		
+		d(printf ("sending gpg our passphrase...\n"));
+		
+		/* send the passphrase to gpg */
+		n = strlen (gpg->passwd);
+		do {
 			do {
-				do {
-					w = write (gpg->secret_fd, gpg->passwd + nwritten, n - nwritten);
-				} while (w == -1 && (errno == EINTR || errno == EAGAIN));
-				
-				if (w > 0)
-					nwritten += w;
-			} while (nwritten < n && w != -1);
+				w = write (gpg->passwd_fd, gpg->passwd + nwritten, n - nwritten);
+			} while (w == -1 && (errno == EINTR || errno == EAGAIN));
 			
-			/* zero and free our passwd buffer */
-			memset (gpg->passwd, 0, n);
-			g_free (gpg->passwd);
-			gpg->passwd = NULL;
-			
-			if (w == -1)
-				goto exception;
-			
-			gpg->send_passwd = FALSE;
-		} else if (gpg->mode == GPG_CTX_MODE_VERIFY) {
-			ssize_t nread, w, nwritten = 0;
-			char buffer[4096];
-			
-			d(printf ("sending gpg our digital signature...\n"));
-			
-			/* write our stream to gpg's sig fd */
-			nread = g_mime_stream_read (gpg->sigstream, buffer, sizeof (buffer));
-			if (nread > 0) {
-				do {
-					do {
-						w = write (gpg->secret_fd, buffer + nwritten, nread - nwritten);
-					} while (w == -1 && (errno == EINTR || errno == EAGAIN));
-					
-					if (w > 0)
-						nwritten += w;
-				} while (nwritten < nread && w != -1);
-				
-				if (w == -1)
-					goto exception;
-			}
-			
-			if (g_mime_stream_eos (gpg->sigstream)) {
-				close (gpg->secret_fd);
-				gpg->secret_fd = -1;
-			}
-		}
+			if (w > 0)
+				nwritten += w;
+		} while (nwritten < n && w != -1);
+		
+		/* zero and free our passwd buffer */
+		memset (gpg->passwd, 0, n);
+		g_free (gpg->passwd);
+		gpg->passwd = NULL;
+		
+		if (w == -1)
+			goto exception;
+		
+		gpg->send_passwd = FALSE;
 	}
 	
 	if ((pfds[GPG_STDIN_FD].revents & (POLLOUT | POLLHUP)) && gpg->istream) {
@@ -1738,7 +1670,7 @@ gpg_sign (GMimeCipherContext *context, const char *userid, GMimeCipherHash hash,
 		diagnostics = gpg_ctx_get_diagnostics (gpg);
 		errno = save;
 		
-		g_set_error_literal (err, GMIME_ERROR, errno, diagnostics);
+		g_set_error (err, GMIME_ERROR, errno, diagnostics);
 		gpg_ctx_free (gpg);
 		
 		return -1;
@@ -1753,6 +1685,35 @@ gpg_sign (GMimeCipherContext *context, const char *userid, GMimeCipherHash hash,
 }
 
 
+static char *
+swrite (GMimeStream *istream)
+{
+	GMimeStream *ostream;
+	char *template;
+	int fd, ret;
+	
+	template = g_build_filename (g_get_tmp_dir (), "gmime-pgp.XXXXXX", NULL);
+	if ((fd = mkstemp (template)) == -1) {
+		g_free (template);
+		return NULL;
+	}
+	
+	ostream = g_mime_stream_fs_new (fd);
+	if ((ret = g_mime_stream_write_to_stream (istream, ostream)) != -1) {
+		if ((ret = g_mime_stream_flush (ostream)) != -1)
+			ret = g_mime_stream_close (ostream);
+	}
+	g_object_unref (ostream);
+	
+	if (ret == -1) {
+		unlink (template);
+		g_free (template);
+		return NULL;
+	}
+	
+	return template;
+}
+
 static GMimeSignatureValidity *
 gpg_verify (GMimeCipherContext *context, GMimeCipherHash hash,
 	    GMimeStream *istream, GMimeStream *sigstream,
@@ -1762,28 +1723,38 @@ gpg_verify (GMimeCipherContext *context, GMimeCipherHash hash,
 	GMimeSignatureValidity *validity;
 	const char *diagnostics;
 	struct _GpgCtx *gpg;
+	char *sigfile = NULL;
 	gboolean valid;
 	
+	if (sigstream != NULL) {
+		/* We are going to verify a detached signature so save
+		   the signature to a temp file. */
+		if (!(sigfile = swrite (sigstream))) {
+			g_set_error (err, GMIME_ERROR, errno,
+				     _("Cannot verify message signature: "
+				       "could not create temp file: %s"),
+				     g_strerror (errno));
+			return NULL;
+		}
+	}
+	
 	gpg = gpg_ctx_new (context->session, ctx->path);
 	gpg_ctx_set_mode (gpg, GPG_CTX_MODE_VERIFY);
-	gpg_ctx_set_sigstream (gpg, sigstream);
-	gpg_ctx_set_istream (gpg, istream);
 	gpg_ctx_set_hash (gpg, hash);
+	gpg_ctx_set_sigfile (gpg, sigfile);
+	gpg_ctx_set_istream (gpg, istream);
 	
 	if (gpg_ctx_op_start (gpg) == -1) {
 		g_set_error (err, GMIME_ERROR, errno,
 			     _("Failed to execute gpg: %s"),
 			     errno ? g_strerror (errno) : _("Unknown"));
-		gpg_ctx_free (gpg);
-		
-		return NULL;
+		goto exception;
 	}
 	
 	while (!gpg_ctx_op_complete (gpg)) {
 		if (gpg_ctx_op_step (gpg, err) == -1) {
 			gpg_ctx_op_cancel (gpg);
-			gpg_ctx_free (gpg);
-			return NULL;
+			goto exception;
 		}
 	}
 	
@@ -1810,7 +1781,23 @@ gpg_verify (GMimeCipherContext *context, GMimeCipherHash hash,
 	
 	gpg_ctx_free (gpg);
 	
+	if (sigfile) {
+		unlink (sigfile);
+		g_free (sigfile);
+	}
+	
 	return validity;
+	
+ exception:
+	
+	gpg_ctx_free (gpg);
+	
+	if (sigfile) {
+		unlink (sigfile);
+		g_free (sigfile);
+	}
+	
+	return NULL;
 }
 
 
@@ -1821,7 +1808,7 @@ gpg_encrypt (GMimeCipherContext *context, gboolean sign, const char *userid,
 {
 	GMimeGpgContext *ctx = (GMimeGpgContext *) context;
 	struct _GpgCtx *gpg;
-	guint i;
+	int i;
 	
 	gpg = gpg_ctx_new (context->session, ctx->path);
 	if (sign)
@@ -1863,7 +1850,7 @@ gpg_encrypt (GMimeCipherContext *context, gboolean sign, const char *userid,
 		diagnostics = gpg_ctx_get_diagnostics (gpg);
 		errno = save;
 		
-		g_set_error_literal (err, GMIME_ERROR, errno, diagnostics);
+		g_set_error (err, GMIME_ERROR, errno, diagnostics);
 		gpg_ctx_free (gpg);
 		
 		return -1;
@@ -1913,7 +1900,7 @@ gpg_decrypt (GMimeCipherContext *context, GMimeStream *istream,
 		diagnostics = gpg_ctx_get_diagnostics (gpg);
 		errno = save;
 		
-		g_set_error_literal (err, GMIME_ERROR, errno, diagnostics);
+		g_set_error (err, GMIME_ERROR, errno, diagnostics);
 		gpg_ctx_free (gpg);
 		
 		return NULL;
@@ -1982,7 +1969,7 @@ gpg_import_keys (GMimeCipherContext *context, GMimeStream *istream, GError **err
 		diagnostics = gpg_ctx_get_diagnostics (gpg);
 		errno = save;
 		
-		g_set_error_literal (err, GMIME_ERROR, errno, diagnostics);
+		g_set_error (err, GMIME_ERROR, errno, diagnostics);
 		gpg_ctx_free (gpg);
 		
 		return -1;
@@ -1998,7 +1985,7 @@ gpg_export_keys (GMimeCipherContext *context, GPtrArray *keys, GMimeStream *ostr
 {
 	GMimeGpgContext *ctx = (GMimeGpgContext *) context;
 	struct _GpgCtx *gpg;
-	guint i;
+	int i;
 	
 	gpg = gpg_ctx_new (context->session, ctx->path);
 	gpg_ctx_set_mode (gpg, GPG_CTX_MODE_EXPORT);
@@ -2035,7 +2022,7 @@ gpg_export_keys (GMimeCipherContext *context, GPtrArray *keys, GMimeStream *ostr
 		diagnostics = gpg_ctx_get_diagnostics (gpg);
 		errno = save;
 		
-		g_set_error_literal (err, GMIME_ERROR, errno, diagnostics);
+		g_set_error (err, GMIME_ERROR, errno, diagnostics);
 		gpg_ctx_free (gpg);
 		
 		return -1;
@@ -2096,7 +2083,7 @@ g_mime_gpg_context_get_always_trust (GMimeGpgContext *ctx)
 /**
  * g_mime_gpg_context_set_always_trust:
  * @ctx: a #GMimeGpgContext
- * @always_trust: always trust flag
+ * @always_trust: always truct flag
  *
  * Sets the @always_trust flag on the gpg context which is used for
  * encryption.
