@@ -1,6 +1,6 @@
 /* -*- Mode: C; tab-width: 8; indent-tabs-mode: t; c-basic-offset: 8 -*- */
 /*  GMime
- *  Copyright (C) 2000-2010 Jeffrey Stedfast
+ *  Copyright (C) 2000-2009 Jeffrey Stedfast
  *
  *  This library is free software; you can redistribute it and/or
  *  modify it under the terms of the GNU Lesser General Public License
@@ -23,16 +23,12 @@
 #include <config.h>
 #endif
 
-#include <stdlib.h>
 #include <sys/types.h>
 #include <sys/stat.h>
 #include <string.h>
-#ifdef HAVE_UNISTD_H
 #include <unistd.h>
-#endif
 #include <fcntl.h>
 #include <errno.h>
-#include <time.h>
 
 #include "gmime-multipart.h"
 #include "gmime-utils.h"
@@ -244,7 +240,7 @@ multipart_write_to_stream (GMimeObject *object, GMimeStream *stream)
 	 * not, then it's a broken multipart and so we don't want to
 	 * alter it or we'll completely break the output) */
 	boundary = g_mime_object_get_content_type_parameter (object, "boundary");
-	if (!boundary && !g_mime_header_list_get_stream (object->headers)) {
+	if (!boundary && !g_mime_header_list_has_raw (object->headers)) {
 		g_mime_multipart_set_boundary (multipart, NULL);
 		boundary = g_mime_object_get_content_type_parameter (object, "boundary");
 	}
@@ -427,9 +423,9 @@ g_mime_multipart_get_postface (GMimeMultipart *multipart)
 static void
 multipart_clear (GMimeMultipart *multipart)
 {
-	guint i;
+	int i;
 	
-	for (i = 0; i < multipart->children->len; i++)
+	for (i = 0; i < (int) multipart->children->len; i++)
 		g_object_unref (multipart->children->pdata[i]);
 	
 	g_ptr_array_set_size (multipart->children, 0);
@@ -482,20 +478,18 @@ ptr_array_insert (GPtrArray *array, guint index, gpointer object)
 	unsigned char *dest, *src;
 	guint n;
 	
-	if (index < array->len) {
-		/* need to shift some items */
-		g_ptr_array_set_size (array, array->len + 1);
-		
+	g_ptr_array_set_size (array, array->len + 1);
+	
+	if (index == array->len) {
+		/* need to move items down */
 		dest = ((unsigned char *) array->pdata) + (sizeof (void *) * (index + 1));
 		src = ((unsigned char *) array->pdata) + (sizeof (void *) * index);
 		n = array->len - index - 1;
 		
 		g_memmove (dest, src, (sizeof (void *) * n));
-		array->pdata[index] = object;
-	} else {
-		/* the easy case */
-		g_ptr_array_add (array, object);
 	}
+	
+	array->pdata[index] = object;
 }
 
 static void
@@ -562,7 +556,7 @@ multipart_remove_at (GMimeMultipart *multipart, int index)
 {
 	GMimeObject *part;
 	
-	if ((guint) index >= multipart->children->len)
+	if (index >= multipart->children->len)
 		return NULL;
 	
 	part = multipart->children->pdata[index];
@@ -598,7 +592,7 @@ multipart_get_part (GMimeMultipart *multipart, int index)
 {
 	GMimeObject *part;
 	
-	if ((guint) index >= multipart->children->len)
+	if (index >= multipart->children->len)
 		return NULL;
 	
 	part = multipart->children->pdata[index];
@@ -629,9 +623,9 @@ g_mime_multipart_get_part (GMimeMultipart *multipart, int index)
 static gboolean
 multipart_contains (GMimeMultipart *multipart, GMimeObject *part)
 {
-	guint i;
+	int i;
 	
-	for (i = 0; i < multipart->children->len; i++) {
+	for (i = 0; i < (int) multipart->children->len; i++) {
 		if (part == (GMimeObject *) multipart->children->pdata[i])
 			return TRUE;
 	}
@@ -663,9 +657,9 @@ g_mime_multipart_contains (GMimeMultipart *multipart, GMimeObject *part)
 static int
 multipart_index_of (GMimeMultipart *multipart, GMimeObject *part)
 {
-	guint i;
+	int i;
 	
-	for (i = 0; i < multipart->children->len; i++) {
+	for (i = 0; i < (int) multipart->children->len; i++) {
 		if (part == (GMimeObject *) multipart->children->pdata[i])
 			return i;
 	}
@@ -720,7 +714,6 @@ g_mime_multipart_get_count (GMimeMultipart *multipart)
 static void
 read_random_pool (unsigned char *buffer, size_t bytes)
 {
-#ifdef HAVE_UNISTD_H
 	size_t nread = 0;
 	ssize_t n;
 	int fd;
@@ -742,14 +735,6 @@ read_random_pool (unsigned char *buffer, size_t bytes)
 	} while (nread < bytes);
 	
 	close (fd);
-#else
-	size_t i;
-	
-	srand ((unsigned int) time (NULL));
-	
-	for (i = 0; i < bytes; i++)
-		buffer[i] = (unsigned char) (rand () % 256);
-#endif
 }
 
 static void
