@@ -24,6 +24,7 @@
 #endif
 
 #include <string.h>
+#include <errno.h>
 
 #include "gmime-stream-buffer.h"
 
@@ -56,10 +57,10 @@ static int stream_flush (GMimeStream *stream);
 static int stream_close (GMimeStream *stream);
 static gboolean stream_eos (GMimeStream *stream);
 static int stream_reset (GMimeStream *stream);
-static off_t stream_seek (GMimeStream *stream, off_t offset, GMimeSeekWhence whence);
-static off_t stream_tell (GMimeStream *stream);
+static gint64 stream_seek (GMimeStream *stream, gint64 offset, GMimeSeekWhence whence);
+static gint64 stream_tell (GMimeStream *stream);
 static ssize_t stream_length (GMimeStream *stream);
-static GMimeStream *stream_substream (GMimeStream *stream, off_t start, off_t end);
+static GMimeStream *stream_substream (GMimeStream *stream, gint64 start, gint64 end);
 
 
 static GMimeStreamClass *parent_class = NULL;
@@ -144,6 +145,11 @@ stream_read (GMimeStream *stream, char *buf, size_t len)
 	ssize_t n, nread = 0;
 	size_t offset;
 	
+	if (buffer->source == NULL) {
+		errno = EBADF;
+		return -1;
+	}
+	
 	switch (buffer->mode) {
 	case GMIME_STREAM_BUFFER_BLOCK_READ:
 		while (len > 0) {
@@ -233,6 +239,11 @@ stream_write (GMimeStream *stream, const char *buf, size_t len)
 	ssize_t n, nwritten = 0;
 	size_t left = len;
 	
+	if (buffer->source == NULL) {
+		errno = EBADF;
+		return -1;
+	}
+	
 	switch (buffer->mode) {
 	case GMIME_STREAM_BUFFER_BLOCK_WRITE:
 		while (left > 0) {
@@ -344,6 +355,9 @@ stream_eos (GMimeStream *stream)
 {
 	GMimeStreamBuffer *buffer = (GMimeStreamBuffer *) stream;
 	
+	if (buffer->source == NULL)
+		return TRUE;
+	
 	if (!g_mime_stream_eos (buffer->source))
 		return FALSE;
 	
@@ -364,6 +378,11 @@ stream_reset (GMimeStream *stream)
 {
 	GMimeStreamBuffer *buffer = (GMimeStreamBuffer *) stream;
 	
+	if (buffer->source == NULL) {
+		errno = EBADF;
+		return -1;
+	}
+	
 	switch (buffer->mode) {
 	case GMIME_STREAM_BUFFER_BLOCK_READ:
 	case GMIME_STREAM_BUFFER_BLOCK_WRITE:
@@ -386,11 +405,11 @@ stream_reset (GMimeStream *stream)
 	return 0;
 }
 
-static off_t
-stream_seek_block_read (GMimeStream *stream, off_t offset, GMimeSeekWhence whence)
+static gint64
+stream_seek_block_read (GMimeStream *stream, gint64 offset, GMimeSeekWhence whence)
 {
 	GMimeStreamBuffer *buffer = (GMimeStreamBuffer *) stream;
-	off_t real;
+	gint64 real;
 	
 	/* convert all seeks into a relative seek (aka SEEK_CUR) */
 	switch (whence) {
@@ -401,8 +420,11 @@ stream_seek_block_read (GMimeStream *stream, off_t offset, GMimeSeekWhence whenc
 		if (stream->position == offset)
 			return stream->position;
 		
-		if (offset < 0)
+		if (offset < 0) {
+			/* not allowed to seek to a negative position */
+			errno = EINVAL;
 			return -1;
+		}
 		
 		offset -= stream->position;
 		break;
@@ -420,12 +442,17 @@ stream_seek_block_read (GMimeStream *stream, off_t offset, GMimeSeekWhence whenc
 			return real;
 		}
 		
-		if (offset > 0)
+		if (offset > 0) {
+			/* not allowed to seek past bound_end */
+			errno = EINVAL;
 			return -1;
+		}
 		
 		offset += stream->bound_end;
 		break;
 	default:
+		/* invalid whence argument */
+		errno = EINVAL;
 		return -1;
 	}
 	
@@ -457,12 +484,12 @@ stream_seek_block_read (GMimeStream *stream, off_t offset, GMimeSeekWhence whenc
 	return real;
 }
 
-static off_t
-stream_seek_cache_read (GMimeStream *stream, off_t offset, GMimeSeekWhence whence)
+static gint64
+stream_seek_cache_read (GMimeStream *stream, gint64 offset, GMimeSeekWhence whence)
 {
 	GMimeStreamBuffer *buffer = (GMimeStreamBuffer *) stream;
 	size_t len, total = 0;
-	off_t pos, real;
+	gint64 pos, real;
 	ssize_t nread;
 	
 	switch (whence) {
@@ -474,16 +501,27 @@ stream_seek_cache_read (GMimeStream *stream, off_t offset, GMimeSeekWhence whenc
 		break;
 	case GMIME_STREAM_SEEK_END:
 		if (stream->bound_end == -1) {
-			real = g_mime_stream_seek (buffer->source, offset, whence);
-			if (real == -1 || real < stream->bound_start)
+			/* we have to do a real seek because the end boundary is unknown */
+			if ((real = g_mime_stream_seek (buffer->source, offset, whence)) == -1)
 				return -1;
+			
+			if (real < stream->bound_start) {
+				/* seek offset out of bounds */
+				errno = EINVAL;
+				return -1;
+			}
 		} else {
 			real = stream->bound_end + offset;
-			if (real > stream->bound_end || real < stream->bound_start)
+			if (real > stream->bound_end || real < stream->bound_start) {
+				/* seek offset out of bounds */
+				errno = EINVAL;
 				return -1;
+			}
 		}
 		break;
 	default:
+		/* invalid whence argument */
+		errno = EINVAL;
 		return -1;
 	}
 	
@@ -518,9 +556,12 @@ stream_seek_cache_read (GMimeStream *stream, off_t offset, GMimeSeekWhence whenc
 		if (total < len) {
 			/* we failed to seek that far so reset our bufptr */
 			buffer->bufptr = buffer->buffer + pos;
+			errno = EINVAL;
 			return -1;
 		}
 	} else if (real < stream->bound_start) {
+		/* seek offset out of bounds */
+		errno = EINVAL;
 		return -1;
 	} else {
 		/* seek our cache pointer backwards */
@@ -532,12 +573,16 @@ stream_seek_cache_read (GMimeStream *stream, off_t offset, GMimeSeekWhence whenc
 	return real;
 }
 
-static off_t
-stream_seek (GMimeStream *stream, off_t offset, GMimeSeekWhence whence)
+static gint64
+stream_seek (GMimeStream *stream, gint64 offset, GMimeSeekWhence whence)
 {
-	/* FIXME: set errno appropriately?? */
 	GMimeStreamBuffer *buffer = (GMimeStreamBuffer *) stream;
-	off_t real;
+	gint64 real;
+	
+	if (buffer->source == NULL) {
+		errno = EBADF;
+		return -1;
+	}
 	
 	switch (buffer->mode) {
 	case GMIME_STREAM_BUFFER_BLOCK_WRITE:
@@ -555,24 +600,40 @@ stream_seek (GMimeStream *stream, off_t offset, GMimeSeekWhence whence)
 	case GMIME_STREAM_BUFFER_CACHE_READ:
 		return stream_seek_cache_read (stream, offset, whence);
 	default:
+		/* invalid whence argument */
+		errno = EINVAL;
 		return -1;
 	}
 }
 
-static off_t
+static gint64
 stream_tell (GMimeStream *stream)
 {
+	GMimeStreamBuffer *buffer = (GMimeStreamBuffer *) stream;
+	
+	if (buffer->source == NULL) {
+		errno = EBADF;
+		return -1;
+	}
+	
 	return stream->position;
 }
 
 static ssize_t
 stream_length (GMimeStream *stream)
 {
-	return g_mime_stream_length (GMIME_STREAM_BUFFER (stream)->source);
+	GMimeStreamBuffer *buffer = (GMimeStreamBuffer *) stream;
+	
+	if (buffer->source == NULL) {
+		errno = EBADF;
+		return -1;
+	}
+	
+	return g_mime_stream_length (buffer->source);
 }
 
 static GMimeStream *
-stream_substream (GMimeStream *stream, off_t start, off_t end)
+stream_substream (GMimeStream *stream, gint64 start, gint64 end)
 {
 	GMimeStreamBuffer *buffer = (GMimeStreamBuffer *) stream;
 	
@@ -592,7 +653,7 @@ stream_substream (GMimeStream *stream, off_t start, off_t end)
  *
  * Creates a new GMimeStreamBuffer object.
  *
- * Returns a new buffer stream with source @source and mode @mode.
+ * Returns: a new buffer stream with source @source and mode @mode.
  **/
 GMimeStream *
 g_mime_stream_buffer_new (GMimeStream *source, GMimeStreamBufferMode mode)
@@ -601,7 +662,7 @@ g_mime_stream_buffer_new (GMimeStream *source, GMimeStreamBufferMode mode)
 	
 	g_return_val_if_fail (GMIME_IS_STREAM (source), NULL);
 	
-	buffer = g_object_new (GMIME_TYPE_STREAM_BUFFER, NULL);
+	buffer = g_object_newv (GMIME_TYPE_STREAM_BUFFER, 0, NULL);
 	
 	buffer->source = source;
 	g_object_ref (source);
@@ -640,11 +701,11 @@ g_mime_stream_buffer_new (GMimeStream *source, GMimeStreamBufferMode mode)
  *
  * Reads in at most one less than @max characters from @stream and
  * stores them into the buffer pointed to by @buf. Reading stops after
- * an EOS or newline (#'\n'). If a newline is read, it is stored into
- * the buffer. A #'\0' is stored after the last character in the
+ * an EOS or newline ('\n'). If a newline is read, it is stored into
+ * the buffer. A '\0' is stored after the last character in the
  * buffer.
  *
- * Returns the number of characters read into @buf on success and -1
+ * Returns: the number of characters read into @buf on success or %-1
  * on fail.
  **/
 ssize_t
